
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <malloc.h>

#ifdef LINUX
#include <signal.h>
#include <unistd.h>
#endif

#include "csearch.h"



uint64_t validate(Problem * p)
{
    unsigned * region_size = (unsigned *) 
                    alloca(p->num_nodes * sizeof(unsigned));
    unsigned i;
    uint64_t value;

    memset(region_size, 0, p->num_nodes * sizeof(unsigned));
    assert(p->num_nodes != 0);

    for (i = 0; i < p->num_nodes; i++) {
        Node * n = &p->nodes[i];

        assert(n->region < p->num_nodes);
        region_size[n->region] += n->size;
        if (region_size[n->region] > p->spm_size) {
            return 0; // invalid
        }
        assert(n->parent <= i);
    }
    assert(p->nodes[0].parent == 0);
    assert(p->nodes[0].region == 0);
   
    if (p->mode == 'P') {
        value = 0;
    } else {
        value = (region_size[0] + p->fixed_cost);
    }

    for (i = 1; i < p->num_nodes; i++) {
        Node * nc = &p->nodes[i];
        Node * np = &p->nodes[nc->parent];

        if (nc->region != np->region) {
            if (p->mode == 'P') {
                value += (uint64_t) nc->edge_weight;
            } else {
                value += (uint64_t) nc->edge_weight *
                    (uint64_t) (p->fixed_cost + region_size[nc->region]);

                if ((p->mode != 'W') && (p->mode != 'w')) {
                    value += (uint64_t) nc->edge_weight *
                        (uint64_t) (p->fixed_cost + region_size[np->region]);
                }
            }
        } 
    }
    for (i = 0; i < p->num_back_edges; i++) {
        Back_Edge * be = &p->back_edges[i];
        Node * h = &p->nodes[be->header];
        Node * f = &p->nodes[be->footer];

        assert ((p->mode == 'W') || (p->mode == 'w'));
        if (h->region != f->region) {
            value += (uint64_t) be->edge_weight *
                (uint64_t) (p->fixed_cost + region_size[h->region]);
        }
    }
    return value;
}

void print_partition(Problem * p, FILE * fd)
{
    int empty_region = 0;
    unsigned region, i;

    fprintf(fd, "[");
    for (region = 0; (!empty_region); region++) {
        empty_region = 1;
        for (i = 0; i < p->num_nodes; i++) {
            Node * n = &p->nodes[i];

            if (n->region == region) {
                if (empty_region) {
                    if (region) {
                        fprintf(fd, ", ");
                    }
                    fprintf(fd, "[");
                    empty_region = 0;
                } else {
                    fprintf(fd, ", ");
                }
                fprintf(fd, "%u", i);
            }
        }
        if (!empty_region) {
            fprintf(fd, "]");
        }
    }
    fprintf(fd, "]\n");
}

Problem * read_problem(FILE * fd)
{
    Problem * p = (Problem *) calloc(1, sizeof(Problem));
    unsigned i;

    if (!p) {
        fprintf(stderr, "Unable to calloc() header\n");
        return 0;
    }

    if ((5 != fscanf(fd, "%u %u %c %u %u\n",
            &p->spm_size, &p->num_nodes, &p->mode, 
            &p->quantum_size, &p->fixed_cost))
    || (!strchr("PCWwE", p->mode))
    || (!p->spm_size) || (p->spm_size > MAX_SPM_SIZE)
    || (!p->quantum_size) || (!p->num_nodes)) {
        fprintf(stderr, "Format error in header\n");
        goto delete_p;
    }

    p->nodes = (Node *) calloc(p->num_nodes, sizeof(Node));
    if (!p->nodes) {
        fprintf(stderr, "Unable to calloc() space for %u nodes.\n", 
                    p->num_nodes);
        goto delete_p;
    }

    p->spm_size = ((p->spm_size + p->quantum_size - 1) / 
                        p->quantum_size) * p->quantum_size;

    for (i = 0; i < p->num_nodes; i++) {
        Node * n = &p->nodes[i];

        if ((4 != fscanf(fd, "%u %u %u %u\n",
            &n->size, &n->parent, &n->edge_weight, &n->region))
        || (!n->size)
        || (n->size > p->spm_size)
        || (!n->edge_weight)) {
            fprintf(stderr, "Format error in node %u\n", i);
            goto delete_np;
        }
        n->size = ((n->size + p->quantum_size - 1) / 
                        p->quantum_size) * p->quantum_size;

        assert (n->size);

        if (i != 0) {
            if (n->parent == i) {
                fprintf(stderr, "Node %u is its own parent (and not root)\n", i);
                goto delete_np;
            } else if (n->parent > i) {
                fprintf(stderr, "Node %u has a parent with a greater number "
                        "(%u)\n", i, n->parent);
                goto delete_np;
            }
        }
    }

    p->num_back_edges = 0;
    if (toupper(p->mode) == 'W') {
        if (1 != fscanf(fd, "%u\n", &p->num_back_edges)) {
            fprintf(stderr, "Format error in header (back edge data)\n");
            goto delete_np;
        }
        p->back_edges = (Back_Edge *) 
                        calloc(p->num_back_edges, sizeof(Back_Edge));
        if (!p->back_edges) {
            fprintf(stderr, "Unable to calloc() space for %u back edges.\n", 
                        p->num_back_edges);
            goto delete_np;
        }
    }

    for (i = 0; i < p->num_back_edges; i++) {
        Back_Edge * be = &p->back_edges[i];
        unsigned pos;

        if ((3 != fscanf(fd, "%u %u %u\n",
                &be->header, &be->footer, &be->edge_weight))
        || (be->header >= be->footer)
        || (be->header >= p->num_nodes)
        || (be->footer >= p->num_nodes)
        || (be->edge_weight == 0)) {
            fprintf(stderr, "Format error in back edge %u.\n", i);
            goto delete_nbp;
        }
        // be->header must be an ancestor of be->footer
        pos = be->footer;
        while ((pos != 0) && (pos != be->header)) {
            pos = p->nodes[pos].parent;
        }
        if (pos != be->header) {
            fprintf(stderr, "Back edge %u: header %u is "
                            "not ancestor of footer %u.\n", 
                                    i, be->header, be->footer);
            goto delete_nbp;
        }
    }
    return p;

delete_nbp:
    free(p->back_edges);
delete_np:
    free(p->nodes);
delete_p:
    free(p);
    return 0;
}

void write_problem(Problem * p, FILE * fd)
{
    uint64_t value;
    unsigned i;

    fprintf(fd, "%u %u %c %u %u\n",
            p->spm_size, p->num_nodes, p->mode, p->quantum_size,
            p->fixed_cost);
    for (i = 0; i < p->num_nodes; i++) {
        Node * n = &p->nodes[i];
        fprintf(fd, "%-8u %-8u %-8u %u\n",
            n->size,
            n->parent,
            n->edge_weight,
            n->region);
    }
    if (toupper(p->mode) == 'W') {
        fprintf(fd, "%u\n", p->num_back_edges);
        for (i = 0; i < p->num_back_edges; i++) {
            Back_Edge * be = &p->back_edges[i];
            fprintf(fd, "%-4u %-4u %u\n",
                be->header, be->footer, be->edge_weight);
        }
    }
    value = validate(p);
    if (value) {
        fprintf(fd, "%1.0f\n", (double) value);
    }
}

void delete_problem(Problem * p)
{
    free(p->back_edges);
    free(p->nodes);
    free(p);
}

#ifdef LINUX
static void alarm_handler(int x)
{
    fprintf(stderr, "solver timeout\n");
    exit(71);
}
#endif

uint64_t csearch(const char * problem, char algorithm, int enable_printf)
{
    FILE * fd;
    Problem * p;
    uint64_t value0, value1; 

    fd = fopen(problem, "rt");
    if (!fd) {
        perror("opening file");
        fprintf(stderr, "Unable to open the problem file.\n");
        return 0;
    }
    p = read_problem(fd);
    fclose(fd);
    if (!p) {
        return 0;
    }
    switch(toupper(algorithm)) {
    case 'X':   value0 = exhaustive_search(p); 
                break;
    case 'L':   value0 = extended_lukes(p);
                break;
#ifdef LINUX
    case 'A':   {
                    struct sigaction sa;
                    sa.sa_handler = alarm_handler;
                    sa.sa_flags = 0;
                    sigaction(SIGALRM, &sa, NULL);
                    alarm(60);
                    value0 = extended_lukes(p);
                    alarm(0);
                }
                break;
#endif
    default:    fprintf(stderr, "Unknown algorithm.\n");
                return 0;
    }
    if (enable_printf) {
        value1 = validate(p);
        fprintf(stdout, "%1.0f %1.0f\n", (double) value0, (double) value1);
        print_partition(p, stdout);
        assert (value0 == value1);
    }
    fd = fopen(problem, "wt");
    if (fd) {
        write_problem(p, fd);
        fclose(fd);
    }
    delete_problem(p);
    return value0;
}





